import io.netty.bootstrap.ServerBootstrap
import io.netty.channel.ChannelFuture
import io.netty.channel.ChannelInitializer
import io.netty.channel.ChannelOption
import io.netty.channel.EventLoopGroup
import io.netty.channel.nio.NioEventLoopGroup
import io.netty.channel.socket.SocketChannel
import io.netty.channel.socket.nio.NioServerSocketChannel
import io.netty.buffer.ByteBuf
import io.netty.channel.ChannelHandlerContext
import io.netty.channel.ChannelInboundHandlerAdapter
import io.netty.util.CharsetUtil
import java.util.Arrays
import io.netty.buffer.Unpooled

class Server(internal var port: Int) {
    internal var concurrentClientCount = 0

    init {
        this.concurrentClientCount = 0
    }

    @Throws(Exception::class)
    fun run() {
        val bossGroup = NioEventLoopGroup()
        val workerGroup = NioEventLoopGroup()
        try {
            val b = ServerBootstrap()
            b.group(bossGroup, workerGroup)
                    .channel(NioServerSocketChannel::class.java)
                    .childHandler(object : ChannelInitializer<SocketChannel>() {
                        @Throws(Exception::class)
                        public override fun initChannel(ch: SocketChannel) {
                            ch.pipeline().addLast(ServerHandler())
                        }
                    })
                    .option(ChannelOption.SO_BACKLOG, 128)
                    .childOption(ChannelOption.SO_KEEPALIVE, true)

            // Bind and start to accept incoming connections.
            val f = b.bind(port).sync()

            // Wait until the server socket is closed.
            // In this example, this does not happen, but you can do that to gracefully
            // shut down your server.
            f.channel().closeFuture().sync()
        } finally {
            workerGroup.shutdownGracefully()
            bossGroup.shutdownGracefully()
        }
    }


    internal inner class ServerHandler : ChannelInboundHandlerAdapter() {
        private var dataForProcessing = ""
        override fun channelActive(ctx: ChannelHandlerContext) {
            concurrentClientCount++
            print("$concurrentClientCount concurrent clients are connected\n")
        }

        override fun channelRead(ctx: ChannelHandlerContext, msg: Any) {
            try {
                val queriesString = dataForProcessing + (msg as ByteBuf).toString(CharsetUtil.UTF_8)
                println("queriesString = $queriesString")
                var queries = queriesString.split(System.lineSeparator()) as MutableList<String>
                println("len(queries) = ${queries.size}")

                // if last query isn't complete then keep it
                if (!queriesString.endsWith(System.lineSeparator())){
                    dataForProcessing = queries.last()
                    queries.removeAt(queries.size - 1)
                    println("Last query was dropped")
                    println("len(queries) = ${queries.size}")
                }
                else dataForProcessing = ""

                for (query in queries) {
                    if (query.isBlank())
                        continue

                    println("Current query = $query")
                    try {
                        val answer = calc(query)
                        println("answer = $answer")
                        ctx.write(Unpooled.copiedBuffer(answer, CharsetUtil.UTF_8))
                    }
                    catch (e: Exception) {
                        println("ERROR")
                        ctx.write(Unpooled.copiedBuffer("ERR\r\n", CharsetUtil.UTF_8))
                    }
                }

                ctx.flush()
            }
            catch (e: Exception){
                ctx.write(Unpooled.copiedBuffer("ERR\r\n", CharsetUtil.UTF_8))
                ctx.flush()
            }
            finally {
                (msg as ByteBuf).release()
            }
        }

        override fun channelInactive(ctx: ChannelHandlerContext) {
            concurrentClientCount--
        }

        fun calc(query: String): String {
            val words = query.trim().split(" ", "\r", "\n")

            var answer = "OK "
            val op = words[1]
            when(op){
                "*" -> answer += words[2].toInt() * words[3].toInt()
                "+" -> answer += words[2].toInt() + words[3].toInt()
                "-" -> answer += words[2].toInt() - words[3].toInt()
                "/" -> answer += words[2].toInt() / words[3].toInt()
                else -> throw IllegalArgumentException()
            }
            answer += "\r\n"

            return answer
        }
    }

    companion object {
        @Throws(Exception::class)
        @JvmStatic
        fun main(args: Array<String>) {
            Server(28563).run()
        }
    }
}




